"""Re-usable utility functions used by this package."""

from __future__ import annotations

import copy
import os

from dataclasses import dataclass
from importlib import resources as impl_resources
from pathlib import Path
from typing import TYPE_CHECKING

import yaml

from ansible_creator.constants import SKIP_DIRS, SKIP_FILES_TYPES


if TYPE_CHECKING:

    from ansible_creator.compat import Traversable
    from ansible_creator.output import Output
    from ansible_creator.templar import Templar
    from ansible_creator.types import TemplateData


PATH_REPLACERS = {
    "project_org": "scm_org",
    "project_repo": "scm_project",
}


@dataclass
class TermFeatures:
    """Terminal features.

    Attributes:
        color: Enable color output.
        links: Enable clickable links.
    """

    color: bool
    links: bool

    def any_enabled(self: TermFeatures) -> bool:
        """Return True if any features are enabled.

        Returns:
            bool: True if any features are enabled.
        """
        return any((self.color, self.links))


def expand_path(path: str) -> Path:
    """Resolve absolute path.

    Args:
        path: Path to expand.

    Returns:
        Expanded absolute path.
    """
    _path = Path(os.path.expandvars(path))
    _path = _path.expanduser()
    return _path.resolve()


@dataclass
class Copier:
    """Configuration for the Copier class.

    Attributes:
        resources: List of resource containers to copy.
        resource_id: The id of the resource to copy.
        dest: The destination path to copy resources to.
        output: An instance of the Output class.
        template_data: A dictionary containing the original data to render templates with.
        allow_overwrite: A list of paths that should be overwritten at destination.
        index: Index of the current resource being copied.
        resource_root: Root path for the resources.
        templar: An instance of the Templar class.
    """

    resources: list[str]
    resource_id: str
    dest: Path
    output: Output
    template_data: TemplateData
    allow_overwrite: list[str] | None = None
    index: int = 0
    resource_root: str = "ansible_creator.resources"
    templar: Templar | None = None

    @property
    def resource(self: Copier) -> str:
        """Return the current resource being copied."""
        return self.resources[self.index]

    def _recursive_copy(  # noqa: C901, PLR0912
        self: Copier,
        root: Traversable,
        template_data: TemplateData,
    ) -> None:
        """Recursively traverses a resource container and copies content to destination.

        Args:
            root: A traversable object representing root of the container to copy.
            template_data: A dictionary containing current data to render templates with.
        """
        self.output.debug(msg=f"current root set to {root}")

        for obj in root.iterdir():
            overwrite = False
            # resource names may have a . but directories use / in the path
            dest_name = str(obj).split(
                self.resource.replace(".", "/") + "/",
                maxsplit=1,
            )[-1]
            dest_path = self.dest / dest_name
            if self.allow_overwrite and (dest_name in self.allow_overwrite):
                overwrite = True
            # replace placeholders in destination path with real values
            for key, val in PATH_REPLACERS.items():
                if key in str(dest_path) and template_data:
                    str_dest_path = str(dest_path)
                    repl_val = getattr(template_data, val)
                    dest_path = Path(str_dest_path.replace(key, repl_val))

            if obj.is_dir():
                if obj.name in SKIP_DIRS:
                    continue
                if not dest_path.exists():
                    dest_path.mkdir(parents=True)

                # recursively copy the directory
                self._recursive_copy(
                    root=obj,
                    template_data=template_data,
                )

            elif obj.is_file():
                if obj.name.split(".")[-1] in SKIP_FILES_TYPES:
                    continue
                if obj.name == "__meta__.yml":
                    continue
                # remove .j2 suffix at destination
                needs_templating = False
                if dest_path.suffix == ".j2":
                    dest_path = dest_path.with_suffix("")
                    needs_templating = True
                dest_file = Path(self.dest) / dest_path
                self.output.debug(msg=f"dest file is {dest_file}")

                # write at destination only if missing or belongs to overwrite list
                if not dest_file.exists() or overwrite:
                    content = obj.read_text(encoding="utf-8")
                    # only render as templates if both of these are provided,
                    # and original file suffix was j2
                    if self.templar and template_data and needs_templating:
                        content = self.templar.render_from_content(
                            template=content,
                            data=template_data,
                        )
                    with dest_file.open("w", encoding="utf-8") as df_handle:
                        df_handle.write(content)

    def _per_container(self: Copier) -> None:
        """Copy files and directories from a possibly nested source to a destination.

        :param copier_config: Configuration for the Copier class.

        :raises CreatorError: if allow_overwrite is not a list.
        """
        self.output.debug(
            msg=f"starting recursive copy with source container '{self.resource}'",
        )
        self.output.debug(msg=f"allow_overwrite set to {self.allow_overwrite}")

        # Cast the template data to not pollute the original
        template_data = copy.deepcopy(self.template_data)

        # Collect and template any resource specific variables
        meta_file = impl_resources.files(f"{self.resource_root}.{self.resource}") / "__meta__.yml"
        try:
            with meta_file.open("r", encoding="utf-8") as meta_fileh:
                self.output.debug(
                    msg=f"loading resource specific vars from {meta_file}",
                )
                meta = yaml.safe_load(meta_fileh.read())
        except FileNotFoundError:
            meta = {}
            self.output.debug(msg="no resource specific vars found")

        found = meta.get(self.resource_id, {})
        for key, value in found.items():
            if value["template"] and self.templar:
                serialized = yaml.dump(value["value"])
                templated = self.templar.render_from_content(
                    template=serialized,
                    data=template_data,
                )
                deserialized = yaml.safe_load(templated)
                setattr(template_data, key, deserialized)
            else:
                setattr(template_data, key, value["value"])

        self._recursive_copy(
            root=impl_resources.files(f"{self.resource_root}.{self.resource}"),
            template_data=template_data,
        )

    def copy_containers(
        self: Copier,
    ) -> None:
        """Copy multiple containers to destination.

        :param copier_config: Configuration for the Copier class.
        """
        for i in range(len(self.resources)):
            self.index = i
            self._per_container()



"""conftest."""

from __future__ import annotations

import os
import subprocess

from subprocess import CalledProcessError, CompletedProcess
from typing import TYPE_CHECKING, Any

import pytest

from ansible_creator.output import Output
from ansible_creator.utils import TermFeatures


if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path


os.environ["HOME"] = "/home/ansible"
os.environ["DEV_WORKSPACE"] = "collections/ansible_collections"


@pytest.fixture()
def cli() -> Callable[[Any], CompletedProcess[str] | CalledProcessError]:
    """Fixture to run CLI commands.

    Returns:
        function: cli_run function.
    """
    return cli_run


@pytest.fixture()
def output(tmp_path: Path) -> Output:
    """Create an Output class object as fixture.

    Args:
        tmp_path: Temporary path.

    Returns:
        Output: Output class object.
    """
    return Output(
        display="text",
        log_file=str(tmp_path) + "ansible-creator.log",
        log_level="notset",
        log_append="false",
        term_features=TermFeatures(color=False, links=False),
        verbosity=0,
    )


def cli_run(args: list[str]) -> CompletedProcess[str] | CalledProcessError:
    """Execute a command using subprocess.

    Args:
        args: Command to run.

    Returns:
        CompletedProcess: CompletedProcess object.
        CalledProcessError: CalledProcessError object.
    """
    updated_env = os.environ.copy()
    # this helps asserting stdout/stderr
    updated_env.update({"LINES": "40", "COLUMNS": "300", "TERM": "xterm-256color"})
    try:
        result = subprocess.run(
            args,
            shell=True,
            capture_output=True,
            check=True,
            text=True,
            env=updated_env,
        )
    except subprocess.CalledProcessError as err:
        return err
    return result



"""Definitions for ansible-creator init action."""

from __future__ import annotations

import shutil

from pathlib import Path
from typing import TYPE_CHECKING

from ansible_creator.exceptions import CreatorError
from ansible_creator.templar import Templar
from ansible_creator.types import TemplateData
from ansible_creator.utils import Copier


if TYPE_CHECKING:

    from ansible_creator.config import Config
    from ansible_creator.output import Output


class Init:
    """Class representing ansible-creator init subcommand."""

    def __init__(
        self: Init,
        config: Config,
    ) -> None:
        """Initialize the init action.

        Args:
            config: App configuration object.
        """
        self._namespace: str = config.namespace
        self._collection_name: str | None = config.collection_name
        self._init_path: Path = Path(config.init_path)
        self._force = config.force
        self._creator_version = config.creator_version
        self._project = config.project
        self._scm_org = config.scm_org
        self._scm_project = config.scm_project
        self._templar = Templar()
        self.output: Output = config.output

    def run(self: Init) -> None:  # noqa: C901
        """Start scaffolding skeleton.

        Raises:
            CreatorError: When computed collection path is an existing directory or file.
        """
        if (
            self._init_path.parts[-2:] == ("collections", "ansible_collections")
            and self._project == "collection"
            and isinstance(self._collection_name, str)
        ):
            self._init_path = self._init_path / self._namespace / self._collection_name

        self.output.debug(msg=f"final collection path set to {self._init_path}")

        # check if init_path already exists
        if self._init_path.exists():
            # init-path exists and is a file
            if self._init_path.is_file():
                msg = f"the path {self._init_path} already exists, but is a file - aborting"
                raise CreatorError(
                    msg,
                )
            if next(self._init_path.iterdir(), None):
                # init-path exists and is not empty, but user did not request --force
                if not self._force:
                    msg = (
                        f"The directory {self._init_path} is not empty.\n"
                        f"You can use --force to re-initialize this directory."
                        f"\nHowever it will delete ALL existing contents in it."
                    )
                    raise CreatorError(msg)

                # user requested --force, re-initializing existing directory
                self.output.warning(
                    f"re-initializing existing directory {self._init_path}",
                )
                try:
                    shutil.rmtree(self._init_path)
                except OSError as e:
                    err = f"failed to remove existing directory {self._init_path}: {e}"
                    raise CreatorError(err) from e

        # if init_path does not exist, create it
        if not self._init_path.exists():
            self.output.debug(msg=f"creating new directory at {self._init_path}")
            self._init_path.mkdir(parents=True)

        common_resources = [
            "common.devcontainer",
            "common.devfile",
            "common.gitignore",
            "common.vscode",
        ]

        if self._project == "collection":
            if not isinstance(self._collection_name, str):
                msg = "Collection name is required when scaffolding a collection."
                raise CreatorError(msg)
            # copy new_collection container to destination, templating files when found
            self.output.debug(msg="started copying collection skeleton to destination")
            template_data = TemplateData(
                namespace=self._namespace,
                collection_name=self._collection_name,
                creator_version=self._creator_version,
            )
            copier = Copier(
                resources=["new_collection", *common_resources],
                resource_id="new_collection",
                dest=self._init_path,
                output=self.output,
                templar=self._templar,
                template_data=template_data,
            )
            copier.copy_containers()

            self.output.note(
                f"collection {self._namespace}.{self._collection_name} "
                f"created at {self._init_path}",
            )

        else:
            self.output.debug(
                msg="started copying ansible-project skeleton to destination",
            )
            if not isinstance(self._scm_org, str) or not isinstance(
                self._scm_project,
                str,
            ):
                msg = (
                    "Parameters 'scm-org' and 'scm-project' are required when "
                    "scaffolding an ansible-project."
                )
                raise CreatorError(msg)

            template_data = TemplateData(
                creator_version=self._creator_version,
                scm_org=self._scm_org,
                scm_project=self._scm_project,
            )

            copier = Copier(
                resources=["ansible_project", *common_resources],
                resource_id="ansible_project",
                dest=self._init_path,
                output=self.output,
                templar=self._templar,
                template_data=template_data,
            )
            copier.copy_containers()

            self.output.note(
                f"ansible project created at {self._init_path}",
            )



"""Unit tests for ansible-creator init."""

from __future__ import annotations

import re
import sys

from collections.abc import Callable
from pathlib import Path
from subprocess import CalledProcessError, CompletedProcess
from typing import Any


cli_type = Callable[[Any], CompletedProcess[str] | CalledProcessError]

CREATOR_BIN = Path(sys.executable).parent / "ansible-creator"


def test_run_help(cli: cli_type) -> None:
    """Test running ansible-creator --help.

    Args:
        cli: cli_run function.
    """
    # Get the path to the current python interpreter
    result = cli(f"{CREATOR_BIN} --help")
    assert result.returncode == 0, (result.stdout, result.stderr)

    assert "Print ansible-creator version and exit." in result.stdout
    assert "The subcommand to invoke." in result.stdout
    assert "Initialize an Ansible Collection." in result.stdout


def test_run_no_subcommand(cli: cli_type) -> None:
    """Test running ansible-creator without subcommand.

    Args:
        cli: cli_run function.
    """
    result = cli(str(CREATOR_BIN))
    assert result.returncode != 0
    assert "the following arguments are required: subcommand" in result.stderr


def test_run_init_no_input(cli: cli_type) -> None:
    """Test running ansible-creator init without any input.

    Args:
        cli: cli_run function.
    """
    result = cli(f"{CREATOR_BIN} init")
    assert result.returncode != 0
    assert (
        "Error: The argument 'collection' is required when scaffolding a collection"
        in result.stderr
    )


def test_run_init_basic(cli: cli_type, tmp_path: Path) -> None:
    """Test running ansible-creator init with empty/non-empty/force.

    Args:
        cli: cli_run function.
        tmp_path: Temporary path.
    """
    final_dest = f"{tmp_path}/collections/ansible_collections"
    cli(f"mkdir -p {final_dest}")

    result = cli(
        f"{CREATOR_BIN} init testorg.testcol --init-path {final_dest}",
    )
    assert result.returncode == 0

    # check stdout
    assert re.search("Note: collection testorg.testcol created at", result.stdout) is not None

    # fail to override existing collection with force=false (default)
    result = cli(
        f"{CREATOR_BIN} init testorg.testcol --init-path {final_dest}",
    )

    assert result.returncode != 0

    # this is required to handle random line breaks in CI, especially with macos runners
    mod_stderr = "".join([line.strip() for line in result.stderr.splitlines()])
    assert (
        re.search(
            rf"Error:\s*The\s*directory\s*{final_dest}/testorg/testcol\s*is\s*not\s*empty.",
            mod_stderr,
        )
        is not None
    )
    assert "You can use --force to re-initialize this directory." in result.stderr
    assert "However it will delete ALL existing contents in it." in result.stderr

    # override existing collection with force=true
    result = cli(f"{CREATOR_BIN} init testorg.testcol --init-path {tmp_path} --force")
    assert result.returncode == 0
    assert re.search("Warning: re-initializing existing directory", result.stdout) is not None



# cspell: ignore dcmp, subdcmp
"""Unit tests for ansible-creator init."""

from __future__ import annotations

import re
import shutil

from filecmp import dircmp
from pathlib import Path
from typing import TypedDict

import pytest

from ansible_creator.config import Config
from ansible_creator.exceptions import CreatorError
from ansible_creator.output import Output
from ansible_creator.subcommands.init import Init
from ansible_creator.utils import TermFeatures
from tests.defaults import FIXTURES_DIR


class ConfigDict(TypedDict):
    """Type hint for Config dictionary.

    Attributes:
        creator_version: The version of the creator.
        output: The output object to use for logging.
        subcommand: The subcommand to execute.
        collection: The name of the collection.
        init_path: Path to initialize the project.
        project: The type of project to scaffold.
        force: Force overwrite of existing directory.
        scm_org: The SCM organization for the project.
        scm_project: The SCM project for the project.
    """

    creator_version: str
    output: Output
    subcommand: str
    collection: str
    init_path: str
    project: str
    force: bool
    scm_org: str | None
    scm_project: str | None


@pytest.fixture(name="cli_args")
def fixture_cli_args(tmp_path: Path, output: Output) -> ConfigDict:
    """Create a dict to use for a Init class object as fixture.

    Args:
        tmp_path: App configuration object.
        output: Output class object.

    Returns:
        dict: Dictionary, partial Init class object.
    """
    return {
        "creator_version": "0.0.1",
        "output": output,
        "subcommand": "init",
        "collection": "testorg.testcol",
        "init_path": str(tmp_path / "testorg" / "testcol"),
        "project": "",
        "force": False,
        "scm_org": "",
        "scm_project": "",
    }


def has_differences(dcmp: dircmp[str], errors: list[str]) -> list[str]:
    """Recursively check for differences in dircmp object.

    Args:
        dcmp: dircmp object.
        errors: List of errors.

    Returns:
        list: List of errors.
    """
    errors.extend([f"Only in {dcmp.left}: {f}" for f in dcmp.left_only])
    errors.extend([f"Only in {dcmp.right}: {f}" for f in dcmp.right_only])
    errors.extend(
        [f"Differing files: {dcmp.left}/{f} {dcmp.right}/{f}" for f in dcmp.diff_files],
    )
    for subdcmp in dcmp.subdirs.values():
        errors = has_differences(subdcmp, errors)
    return errors


def test_run_success_for_collection(
    capsys: pytest.CaptureFixture[str],
    tmp_path: Path,
    cli_args: ConfigDict,
) -> None:
    """Test Init.run().

    Args:
        capsys: Pytest fixture to capture stdout and stderr.
        tmp_path: Temporary directory path.
        cli_args: Dictionary, partial Init class object.
    """
    cli_args["project"] = "collection"
    init = Init(
        Config(**cli_args),
    )
    init.run()
    result = capsys.readouterr().out

    # check stdout
    assert re.search("Note: collection testorg.testcol created", result) is not None

    # recursively assert files created
    cmp = dircmp(str(tmp_path), str(FIXTURES_DIR / "collection"))
    diff = has_differences(dcmp=cmp, errors=[])
    assert diff == [], diff

    # fail to override existing collection with force=false (default)
    fail_msg = (
        f"The directory {tmp_path}/testorg/testcol is not empty."
        "\nYou can use --force to re-initialize this directory."
        "\nHowever it will delete ALL existing contents in it."
    )
    with pytest.raises(CreatorError, match=fail_msg):
        init.run()

    # override existing collection with force=true
    cli_args["force"] = True
    init = Init(
        Config(**cli_args),
    )
    init.run()
    result = capsys.readouterr().out
    assert re.search("Warning: re-initializing existing directory", result) is not None, result


def test_run_success_ansible_project(
    capsys: pytest.CaptureFixture[str],
    tmp_path: Path,
    cli_args: ConfigDict,
) -> None:
    """Test Init.run().

    Successfully create new ansible-project

    Args:
        capsys: Pytest fixture to capture stdout and stderr.
        tmp_path: Temporary directory path.
        cli_args: Dictionary, partial Init class object.
    """
    cli_args["collection"] = ""
    cli_args["project"] = "ansible-project"
    cli_args["init_path"] = str(tmp_path / "new_project")
    cli_args["scm_org"] = "weather"
    cli_args["scm_project"] = "demo"
    init = Init(
        Config(**cli_args),
    )
    init.run()
    result = capsys.readouterr().out

    # check stdout
    assert re.search("Note: ansible project created", result) is not None

    # recursively assert files created
    cmp = dircmp(
        str(tmp_path / "new_project"),
        str(FIXTURES_DIR / "project" / "ansible_project"),
    )
    diff = has_differences(dcmp=cmp, errors=[])
    assert diff == [], diff

    # fail to override existing ansible-project directory with force=false (default)
    fail_msg = (
        f"The directory {tmp_path}/new_project is not empty."
        "\nYou can use --force to re-initialize this directory."
        "\nHowever it will delete ALL existing contents in it."
    )
    with pytest.raises(CreatorError, match=fail_msg):
        init.run()

    # override existing ansible-project directory with force=true
    cli_args["force"] = True
    init = Init(
        Config(**cli_args),
    )
    init.run()
    result = capsys.readouterr().out
    assert re.search("Warning: re-initializing existing directory", result) is not None, result


def test_run_success_collections_alt_dir(
    tmp_path: Path,
    capsys: pytest.CaptureFixture[str],
    cli_args: ConfigDict,
) -> None:
    """Test Init.run() when init_path ends with "collections" / "ansible_collections.

    Successfully create new collection

    Args:
        tmp_path: Temporary directory path.
        capsys: Pytest fixture to capture stdout and stderr.
        cli_args: Dictionary, partial Init class object.
    """
    cli_args["project"] = "collection"
    cli_args["init_path"] = str(tmp_path / "collections" / "ansible_collections")
    final_path = Path(cli_args["init_path"]) / "testorg" / "testcol"
    init = Init(
        Config(**cli_args),
    )
    init.run()
    result = capsys.readouterr().out

    # this is required to handle random line breaks in CI, especially with macos runners
    mod_result = "".join([line.strip() for line in result.splitlines()])

    assert (
        re.search(
            rf"Note:\s*collection\s*testorg.testcol\s*created\s*at\s*{final_path}",
            mod_result,
        )
        is not None
    )


def test_error_1(
    tmp_path: Path,
    cli_args: ConfigDict,
) -> None:
    """Test Init.run().

    Validation for: ansible-creator init --project=ansible-project

    Args:
        tmp_path: Temporary directory path.
        cli_args: Dictionary, partial Init class object.
    """
    cli_args["collection"] = ""
    cli_args["project"] = "ansible-project"
    cli_args["init_path"] = str(tmp_path / "new_project")
    cli_args["scm_org"] = None
    cli_args["scm_project"] = None
    fail_msg = (
        "Parameters 'scm-org' and 'scm-project' are required when "
        "scaffolding an ansible-project."
    )
    with pytest.raises(CreatorError, match=fail_msg):
        Init(Config(**cli_args))


def test_error_2(
    cli_args: ConfigDict,
) -> None:
    """Test Init.run().

    Validation for: ansible-creator init

    Args:
        cli_args: Dictionary, partial Init class object.
    """
    cli_args["collection"] = ""
    cli_args["project"] = "collection"
    cli_args["init_path"] = ""
    cli_args["scm_org"] = ""
    cli_args["scm_project"] = ""
    fail_msg = "The argument 'collection' is required when scaffolding a collection."
    with pytest.raises(CreatorError, match=fail_msg):
        Init(Config(**cli_args))


def test_warning(
    capsys: pytest.CaptureFixture[str],
    tmp_path: Path,
    cli_args: ConfigDict,
) -> None:
    """Test Init.run().

    Validation for: ansible-creator init testorg.testname --scm-org=weather
    --scm-project=demo --project=collection

    Args:
        capsys: Pytest fixture to capture stdout and stderr.
        tmp_path: Temporary directory path.
        cli_args: Dictionary, partial Init class object.
    """
    cli_args["collection"] = "testorg.testname"
    cli_args["project"] = "collection"
    cli_args["init_path"] = str(tmp_path / "testorg" / "testcol")
    cli_args["scm_org"] = "weather"
    cli_args["scm_project"] = "demo"
    init = Init(
        Config(**cli_args),
    )
    init.run()
    result = capsys.readouterr().out

    # this is required to handle random line breaks in CI, especially with macos runners
    mod_result = "".join([line.strip() for line in result.splitlines()])
    assert (
        re.search(
            r"Warning:\s*The parameters\s*'scm-org'\s*and\s*'scm-project'"
            r"\s*have\s*no\s*effect\s*when\s*project\s*is\s*not\s*set\s*to\s*ansible-project",
            mod_result,
        )
        is not None
    )


def test_collection_name_char_error(
    cli_args: ConfigDict,
) -> None:
    """Test a collection's name for disallowed characters.

    Validation for: ansible-creator init <collection>

    Args:
        cli_args: Dictionary, partial Init class object.
    """
    cli_args["collection"] = "BAD.NAME"
    cli_args["project"] = "collection"
    cli_args["init_path"] = ""
    cli_args["scm_org"] = ""
    cli_args["scm_project"] = ""
    fail_msg = (
        "Collection name can only contain lower case letters, underscores, and numbers"
        " and cannot begin with an underscore."
    )
    with pytest.raises(CreatorError, match=fail_msg):
        Init(Config(**cli_args))


def test_collection_name_length_error(
    cli_args: ConfigDict,
) -> None:
    """Test a collection's name for length greater than 2 characters.

    Validation for: ansible-creator init <collection>

    Args:
        cli_args: Dictionary, partial Init class object.
    """
    cli_args["collection"] = "na.co"
    cli_args["project"] = "collection"
    cli_args["init_path"] = ""
    cli_args["scm_org"] = ""
    cli_args["scm_project"] = ""
    fail_msg = "Collection namespace and name must be longer than 2 characters."
    with pytest.raises(CreatorError, match=fail_msg):
        Init(Config(**cli_args))


def test_delete_error(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    """Test a remove fails gracefully.

    Args:
        monkeypatch: Pytest monkeypatch fixture.
        tmp_path: Temporary directory path.
    """
    (tmp_path / "file.txt").touch()

    init = Init(
        Config(
            creator_version="0.0.1",
            force=True,
            subcommand="init",
            collection="testorg.testcol",
            init_path=str(tmp_path),
            output=Output(
                log_file=str(tmp_path / "log.log"),
                log_level="DEBUG",
                log_append="false",
                term_features=TermFeatures(color=False, links=False),
                verbosity=0,
            ),
        ),
    )

    err = "Test thrown error"

    def rmtree(path: Path) -> None:  # noqa: ARG001
        raise OSError(err)

    monkeypatch.setattr(shutil, "rmtree", rmtree)

    with pytest.raises(CreatorError, match=err) as exc_info:
        init.run()
    assert "failed to remove existing directory" in str(exc_info.value)


def test_is_file_error(tmp_path: Path) -> None:
    """Test a file dest fails gracefully.

    Args:
        tmp_path: Temporary directory path.
    """
    file = tmp_path / "file.txt"
    file.touch()
    init = Init(
        Config(
            creator_version="0.0.1",
            force=True,
            subcommand="init",
            collection="testorg.testcol",
            init_path=str(file),
            output=Output(
                log_file=str(tmp_path / "log.log"),
                log_level="DEBUG",
                log_append="false",
                term_features=TermFeatures(color=False, links=False),
                verbosity=0,
            ),
        ),
    )

    with pytest.raises(CreatorError) as exc_info:
        init.run()
    assert "but is a file" in str(exc_info.value)


def test_collection_name_not_set(output: Output, tmp_path: Path) -> None:
    """Although it shouldn't happen, test when collection name is not set.

    Args:
        output: Output class object.
        tmp_path: Temporary directory path.
    """

    class FakeConfig:
        """Fake Config class, ours protects from this error.

        Attributes:
            collection: The name of the collection.
            collection_name: The name of the collection.
            creator_version: The version of the creator.
            force: Force overwrite of existing directory.
            init_path: Path to initialize the project.
            project: The type of project to scaffold.
            output: The output object to use for logging.
            namespace: The namespace for the collection.
            scm_org: The SCM organization for the project.
            scm_project: The SCM project for the project.
            subcommand: The subcommand to execute.
        """

        collection: str = "foo.bar"
        collection_name: str | None = None
        creator_version: str = "0.0.1"
        force: bool = False
        init_path: Path = tmp_path
        project: str = "collection"
        output: Output
        namespace: str | None = None
        scm_org: str | None = None
        scm_project: str | None = None
        subcommand: str = "init"

    config = FakeConfig()
    config.output = output

    expected = "Collection name is required when scaffolding a collection."
    with pytest.raises(CreatorError, match=expected):
        Init(config=config).run()  # type: ignore[arg-type]


def test_scm_vals_not_set(output: Output, tmp_path: Path) -> None:
    """Although it shouldn't happen, test when scm_org or scm_project not set.

    Args:
        output: Output class object.
        tmp_path: Temporary directory path.
    """

    class FakeConfig:
        """Fake Config class, ours protects from this error.

        Attributes:
            collection: The name of the collection.
            collection_name: The name of the collection.
            creator_version: The version of the creator.
            force: Force overwrite of existing directory.
            init_path: Path to initialize the project.
            project: The type of project to scaffold.
            output: The output object to use for logging.
            namespace: The namespace for the collection.
            scm_org: The SCM organization for the project.
            scm_project: The SCM project for the project.
            subcommand: The subcommand to execute.
        """

        collection: str = "foo.bar"
        collection_name: str | None = None
        creator_version: str = "0.0.1"
        force: bool = False
        init_path: Path = tmp_path
        project: str = "ansible_project"
        output: Output
        namespace: str | None = None
        scm_org: str | None = None
        scm_project: str | None = None
        subcommand: str = "init"

    config = FakeConfig()
    config.output = output

    expected = (
        "Parameters 'scm-org' and 'scm-project'"
        " are required when scaffolding an ansible-project."
    )

    with pytest.raises(CreatorError, match=expected):
        Init(config=config).run()  # type: ignore[arg-type]



"""Definition of constants for this package."""

GLOBAL_TEMPLATE_VARS = {
    "DEV_CONTAINER_IMAGE": "ghcr.io/ansible/community-ansible-dev-tools:latest",
    "DEV_FILE_IMAGE": "ghcr.io/ansible/ansible-workspace-env-reference:latest",
    "RECOMMENDED_EXTENSIONS": ["redhat.ansible", "redhat.vscode-redhat-account"],
}

MIN_COLLECTION_NAME_LEN = 2

# directory names that will be skipped in any resource
SKIP_DIRS = ("__pycache__",)
# file types that will be skipped in any resource
SKIP_FILES_TYPES = (".pyc",)



"""A home for shared types."""

from collections.abc import Sequence
from dataclasses import dataclass, field

from ansible_creator.constants import GLOBAL_TEMPLATE_VARS


@dataclass
class TemplateData:
    """Dataclass representing the template data.

    Attributes:
        additions: A dictionary containing additional data to add to the gitignore.
        collection_name: The name of the collection.
        creator_version: The version of the creator.
        dev_container_image: The devcontainer image.
        dev_file_image: The devfile image.
        namespace: The namespace of the collection.
        recommended_extensions: A list of recommended VsCode extensions.
        scm_org: The organization of the source control management.
        scm_project: The project of the source control management.
    """

    additions: dict[str, dict[str, dict[str, str | bool]]] = field(default_factory=dict)
    collection_name: str = ""
    creator_version: str = ""
    dev_container_image: Sequence[str] = GLOBAL_TEMPLATE_VARS["DEV_CONTAINER_IMAGE"]
    dev_file_image: Sequence[str] = GLOBAL_TEMPLATE_VARS["DEV_FILE_IMAGE"]
    namespace: str = ""
    recommended_extensions: Sequence[str] = field(
        default_factory=lambda: GLOBAL_TEMPLATE_VARS["RECOMMENDED_EXTENSIONS"],
    )
    scm_org: str = ""
    scm_project: str = ""



"""Test the utils module."""

from pathlib import Path

from ansible_creator.utils import expand_path


def test_expand_path() -> None:
    """Test expand_path utils."""
    home = Path.home().resolve()
    expected = home / "collections/ansible_collections/namespace/collection"
    assert expand_path("~/$DEV_WORKSPACE/namespace/collection") == expected
    assert expand_path("~") == home
    assert expand_path("foo") == Path.cwd() / "foo"
    assert expand_path("$HOME") == home
    assert expand_path("~/$HOME") == Path(f"{home}/{Path.home()}")



"""Basic unit tests for ansible-creator."""

from __future__ import annotations

import re
import sys

from pathlib import Path

import pytest

from ansible_creator.cli import Cli
from ansible_creator.config import Config
from ansible_creator.output import Output
from ansible_creator.utils import TermFeatures, expand_path


def test_configuration_class(output: Output) -> None:
    """Test Config() dataclass post_init.

    Args:
        output: Output dataclass object.
    """
    app_config = Config(
        creator_version="0.0.1",
        subcommand="init",
        collection="testorg.testcol",
        init_path="$HOME",
        output=output,
    )
    assert app_config.namespace == "testorg"
    assert app_config.collection_name == "testcol"
    linux_path = Path("/home/ansible")
    mac_os_path = Path("/System/Volumes/Data/home/ansible")
    assert app_config.init_path in [linux_path, mac_os_path]


@pytest.mark.parametrize(
    argnames=["sysargs", "expected"],
    argvalues=[
        [
            ["ansible-creator", "init", "testorg.testcol"],
            {
                "subcommand": "init",
                "no_ansi": False,
                "log_file": str(Path.cwd() / "ansible-creator.log"),
                "log_level": "notset",
                "log_append": "true",
                "json": False,
                "verbose": 0,
                "collection": "testorg.testcol",
                "init_path": "./",
                "force": False,
                "project": "collection",  # default value
                "scm_org": None,
                "scm_project": None,
            },
        ],
        [
            [
                "ansible-creator",
                "init",
                "--project=ansible-project",
                "--init-path=/home/ansible/my-ansible-project",
            ],
            {
                "subcommand": "init",
                "no_ansi": False,
                "log_file": str(Path.cwd() / "ansible-creator.log"),
                "log_level": "notset",
                "log_append": "true",
                "json": False,
                "verbose": 0,
                "collection": None,
                "init_path": "/home/ansible/my-ansible-project",
                "force": False,
                "project": "ansible-project",
                "scm_org": None,
                "scm_project": None,
            },
        ],
        [
            [
                "ansible-creator",
                "init",
                "testorg.testcol",
                "--init-path=/home/ansible",
                "-vvv",
                "--json",
                "--no-ansi",
                "--la=false",
                "--lf=test.log",
                "--ll=debug",
                "--force",
            ],
            {
                "subcommand": "init",
                "no_ansi": True,
                "log_file": "test.log",
                "log_level": "debug",
                "log_append": "false",
                "json": True,
                "verbose": 3,
                "collection": "testorg.testcol",
                "init_path": "/home/ansible",
                "force": True,
                "project": "collection",  # default value
                "scm_org": None,
                "scm_project": None,
            },
        ],
        [
            [
                "ansible-creator",
                "init",
                "--project=ansible-project",
                "--scm-org=weather",
                "--scm-project=demo",
                "--init-path=/home/ansible/my-ansible-project",
                "-vvv",
                "--json",
                "--no-ansi",
                "--la=false",
                "--lf=test.log",
                "--ll=debug",
                "--force",
            ],
            {
                "subcommand": "init",
                "no_ansi": True,
                "log_file": "test.log",
                "log_level": "debug",
                "log_append": "false",
                "json": True,
                "verbose": 3,
                "collection": None,
                "init_path": "/home/ansible/my-ansible-project",
                "force": True,
                "project": "ansible-project",
                "scm_org": "weather",
                "scm_project": "demo",
            },
        ],
    ],
)
def test_cli_parser(
    monkeypatch: pytest.MonkeyPatch,
    sysargs: list[str],
    expected: dict[str, str | bool | None],
) -> None:
    """Test CLI args parsing.

    Args:
        monkeypatch: Pytest monkeypatch fixture.
        sysargs: List of CLI arguments.
        expected: Expected values for the parsed CLI arguments.
    """
    monkeypatch.setattr("sys.argv", sysargs)
    assert vars(Cli().parse_args()) == expected


def test_missing_j2(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test missing Jinja2.

    Args:
        monkeypatch: Pytest monkeypatch fixture.
    """
    fail_msg = (
        "jinja2 is required but does not appear to be installed."
        "It can be installed using `pip install jinja2`"
    )

    monkeypatch.setattr("sys.path", [])
    monkeypatch.delitem(sys.modules, "jinja2", raising=False)
    monkeypatch.delitem(sys.modules, "ansible_creator.templar", raising=False)

    import ansible_creator.templar

    assert ansible_creator.templar.HAS_JINJA2 is False
    with pytest.raises(ImportError, match=fail_msg):
        ansible_creator.templar.Templar()


def test_cli_init_output(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test CLI init_output method.

    Args:
        monkeypatch: Pytest monkeypatch fixture.
    """
    sysargs = [
        "ansible-creator",
        "init",
        "testorg.testcol",
        "--init-path=/home/ansible",
        "-vvv",
        "--json",
        "--no-ansi",
        "--la=false",
        "--lf=test.log",
        "--ll=debug",
        "--force",
    ]
    output = Output(
        log_append="false",
        log_file=str(expand_path("test.log")),
        log_level="debug",
        term_features=TermFeatures(color=False, links=False),
        verbosity=3,
        display="json",
    )

    monkeypatch.setattr("sys.argv", sysargs)
    cli = Cli()
    cli.init_output()
    assert vars(cli.output) == vars(output)


def test_cli_main(
    capsys: pytest.CaptureFixture[str],
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test CLI main method.

    Args:
        capsys: Pytest capsys fixture.
        tmp_path: Temporary path.
        monkeypatch: Pytest monkeypatch fixture.
    """
    sysargs = [
        "ansible-creator",
        "init",
        "testns.testcol",
        f"--init-path={tmp_path}/testns/testcol",
        "--json",
        "--force",
        "-vvv",
    ]

    monkeypatch.setattr("sys.argv", sysargs)
    cli = Cli()
    cli.init_output()
    cli.run()

    result = capsys.readouterr().out
    # check stdout
    assert re.search("collection testns.testcol created", result) is not None



"""A hello-world filter plugin in testorg.testcol."""

from __future__ import absolute_import, annotations, division, print_function


__metaclass__ = type  # pylint: disable=C0103

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from typing import Callable


DOCUMENTATION = """
    name: hello_world
    author: Testorg Testcol
    version_added: "1.0.0"
    short_description: Demo filter plugin that returns a Hello message.
    description:
      - This is a demo filter plugin designed to return Hello message.
    options:
      name:
        description: Value specified here is appended to the Hello message.
        type: str
"""

EXAMPLES = """
# hello_world filter example

- name: Display a hello message
  ansible.builtin.debug:
    msg: "{{ 'ansible-creator' | testorg.testcol.hello_world }}"
"""


def _hello_world(name: str) -> str:
    """Returns Hello message.

    Args:
        name: The name to greet.

    Returns:
        str: The greeting message.
    """
    return "Hello, " + name


class FilterModule:
    """filter plugin."""

    def filters(self: FilterModule) -> dict[str, Callable[[str], str]]:
        """Map filter plugin names to their functions.

        Returns:
            dict: The filter plugin functions.
        """
        return {"hello_world": _hello_world}
